require 'fileutils'
require 'optparse'
require 'set'


FlickRawOptions = {}

$fileutils = FileUtils::Verbose

# TODO: detect "photo unavailable" and don't download image
UNAVAILABLE_URL = 'http://l.yimg.com/g/images/photo_unavailable.gif'

# TODO: keep track of failures

module Squarepusher
  
  class << self
    
    def describe_photoset(pset)
      "#{pset.id} '#{pset.title}'"
    end
    
    def normalize(name)
      name.gsub(/[^a-zA-Z0-9_+\.=-]+/, '-')
    end

    def owner_page_url(p)
      # TODO: maybe abstract this to allow ownername to be passed in as a parameter
      if p.respond_to?(:owner)
        owner = p.owner
        if owner.respond_to?(:username)
          ownername = p.owner.username
        else
          ownername = owner
        end
      elsif p.respond_to?(:ownername)
        # I want to remove this since the ownername is the human readable form and not the owner id,
        # so isn't always safe to build a URL with.  but for photoset.getList the only way to get the owner name
        # is by specifying the extra "owner_name" - need to find a way to get the owner id
        ownername = p.ownername
      else
        raise Exception.new("photo does not contain 'owner' attribute: #{p.inspect}")
      end
      
      "http://flickr.com/photos/#{ownername}"
    end
    
    def photo_page_url(p)
      owner_page_url = Squarepusher::owner_page_url(p)
      "#{owner_page_url}/#{p.id}"
    end
    
    def sizes
       return [:original, :large, :medium_640, :medium_500, :small, :thumb, :small_square]
    end
    
    def url_for(photo, size)
      case size
          when :original
            FlickRaw.url_o(photo)
          when :large
            FlickRaw.url_b(photo)
          when :medium_640
            FlickRaw.url_z(photo)
          when :medium_500
            FlickRaw.url(photo)
          when :small
            FlickRaw.url_m(photo)
          when :thumb
            FlickRaw.url_t(photo)
          when :small_square
            FlickRaw.url_s(photo)
      end
    end
        
  end

  class Client
    
    class << self
      
      def from_config_path(path, args={})
        open(path) do |stream|
          from_config_stream(stream, args)
        end
      end
    
      def from_config_stream(stream, args={})
        key, secret, token, token_secret = parse_config_stream(stream)
        Squarepusher::Client.new(key, secret, token, token_secret, args)
      end
      
      def parse_config_stream(stream)
        parsed = YAML.load(stream)
        if not parsed.respond_to?(:has_key?)
          raise Exception.new("invalid YAML in #{path}; parsed #{parsed.inspect}")
        end
        puts "parsed config: #{parsed.inspect}"
        required_keys = ['key', 'secret', 'token', 'token_secret']
        missing_keys = []
        vals = []
        required_keys.each do |k|
          v = parsed.fetch(k, nil)
          if not v.nil?
            vals << v
          else
            missing_keys << k 
          end
        end
        if not missing_keys.empty?
          raise Exception.new("missing keys in #{path}: #{missing_keys.inspect}")
        end
        vals
      end
      
    end # end class << self
  
    # optional items for args:
    # :overwrite - overwrite local files
    # :verbose - verbose logging
    # :privacy - privacy level for requests
    # :event_listener - listener object that must provide "event" method
    def initialize(key, secret, token, token_secret, args={})
      @overwrite = args[:overwrite] || false
      @verbose = args[:verbose] || false
      @privacy = args[:privacy] || 5

      @event_listener = args[:event_listener] || nil

      if not @event_listener.nil?
        if not @event_listener.respond_to?(:event)
          puts "ignoring event listener #{@event_listener}; does not provide :event method"
          @event_listener = nil
        else
          puts "using event listener #{@event_listener}"
        end
      end
      
      FlickRawOptions['timeout'] = args[:timeout] || 5
      
      require 'flickraw'
      
      FlickRaw.api_key = key
      FlickRaw.shared_secret = secret
      
      # flickr.auth.checkToken(:auth_token => token)
      flickr.access_token = token
      flickr.access_secret = token_secret
      
      @size = args[:size] || :small_square
    end

    def get_photo(photo_id)
      flickr.photos.getInfo(:photo_id => photo_id)
    end
        
    def get_photoset(pset_id)
      flickr.photosets.getInfo(:photoset_id => pset_id)
    end
    
    def get_photos_for_set(pset_id, params = {})
      params[:photoset_id] = pset_id
      flickr.photosets.getPhotos(params)
    end
    
    def each_photoset
      flickr.photosets.getList.each do |pset|
        yield pset
      end
    end
    
    def search_by_tags(user_id, tags, args={})
      args[:user_id] = user_id
      args[:tags] = tags
      return flickr.photos.search(args)
    end
    
    def random_contact
      contact_hash = flickr.contacts.getList().to_hash()
      pages = contact_hash["pages"].to_i
      puts "WARNING - #{pages} pages of contacts found - only using first one" if pages > 1
      contacts = contact_hash["contact"]
      contacts[rand(contacts.size)]
    end

    def random_favorite_of_mine(args={})
      faves = flickr.favorites.getList(args)
      raise Exception.new("no favorites found") if faves.size == 0

      faves_hash = faves.to_hash
      # puts "hash: #{faves_hash.inspect}"
      photos = faves_hash["photo"]
      pages = faves_hash["pages"]

      fire_event(:domain => :flickr, :user => :self, :favorite_page_count => pages)

      choice = photos[rand(photos.size)]

      choice
    end

    def random_favorite_of_contact(contact_id, args={})
      args[:user_id] = contact_id
      faves = flickr.favorites.getPublicList(args)
      raise Exception.new("no favorites found") if faves.size == 0

      faves_hash = faves.to_hash
      # puts "hash: #{faves_hash.inspect}"
      photos = faves_hash["photo"]
      pages = faves_hash["pages"]

      fire_event(:domain => :flickr, :user => contact_id, :favorite_page_count => pages)

      choice = photos[rand(photos.size)]

      choice
    end

    def download_favorites(output_dir, page=1, &photo_call)
      puts "page: #{page}"
      count = 0
      flickr.favorites.getList(:page => page, :extras => 'owner_name, original_format').each do |p|
        ext = if p.respond_to?(:original_format) then p.originalformat else "jpg" end
        path = "#{output_dir}/#{p.owner}-#{p.id}.#{ext}"
        count += 1
        next if not @overwrite and File.exists?(path)
        u = Squarepusher.url_for(p, :medium_640)
        download_image(u, path)
        if block_given? then photo_call.call(p, u, path) end
      end
      puts "count: #{count}"
      if count == 100 then download_favorites(output_dir, page+1, &photo_call) end
    end
    
    def download_photoset(photoset, output_dir, args={})
      size = args[:size] || @size

      pset_normalized_title = Squarepusher.normalize(photoset.title)
      dirname = pset_normalized_title
      set_dir = File.join(output_dir, dirname)
      $fileutils.mkdir_p set_dir
      
      # contains the following keys
      # error
      # exists      
      results = {}
      local_paths = []
      results[:local_photoset_dir] = set_dir
      results[:local_paths] = local_paths

      # handles socket timeout when loading photoset info
      status, photoset_result = handle_error { flickr.photosets.getPhotos(:photoset_id => photoset.id,
        :privacy_filter => @privacy,
        :extras => "original_format,url_o,date_taken")["photo"] }
      if status == :error
        error photoset_result
        results[status] = 1
        return results
      else
        photos = photoset_result
      end
      
      photos.each do |p|
        # removes all non digits from date
        date_taken = p["datetaken"].gsub(/[^\d]+/, '')

        name = "#{date_taken}-#{p.id}-#{pset_normalized_title}"
        if not p.title.empty?
          normalized_title = Squarepusher.normalize(p.title)
          name << "-#{normalized_title}"
        end
        
        url = Squarepusher.url_for(p, size)
        
        # TODO: only add .jpg suffix if it's not in the image name already
        path = File.join(set_dir, name)
        path << ".jpg" if not path =~ /.jpg$/
        
        if not @overwrite and File.exists?(path)
          debug "#{path} exists; skipping"
          result = :exists
        else
          result = download_image(url, path)
        end
        
        if results.has_key?(result)
          results[result] = results[result] + 1
        else
          results[result] = 1
        end

        if result == :success || result == :exists
          local_paths << path
        end
      end
      results
    end
 
    # syncs local directory with flickr photoset
    # meaning, if a file exists locally that is no longer in the set, it will be deleted
    def sync_photoset(photoset, output_dir, args={})
      results = download_photoset(photoset, output_dir, args)
      local_pset_dir = results[:local_photoset_dir]

      # converts list to set for faster lookups
      local_paths = Set.new(results[:local_paths])

      Dir.entries(local_pset_dir).collect { |e| "#{local_pset_dir}/#{e}"}.select { |e| File.file?(e) }.each do |e|
        if not local_paths.include?(e)
          # puts "deleting #{e}"
          $fileutils.rm e
        end
      end
    end

    private
    
      def fire_event(args={})
        @event_listener.event(args) unless @event_listener.nil?
      end

      # was originally sending this to stderr, but the difference in flushing between
      # stdout and stderr was creating logging confusion
      def error(msg)
        puts "ERROR: #{msg}"
      end
    
      def debug(msg)
        puts msg if @verbose
      end

      def info(msg)
        puts msg
      end
    
      def download_image(url, path, args={})
        if url == UNAVAILABLE_URL
          redirects = args[:redirects]
          msg = "#{url} unavailable"
          msg << "; redirects: #{redirects}" if not redirects.nil?
          debug msg
          return :unavailable
        end
        
        result = :unknown
        uri = URI.parse(url)
        begin
          Net::HTTP.start(uri.host, uri.port) do |http|
            full_path = uri.request_uri
            response = http.get(full_path)
      
            # puts response.inspect
            case response
              when Net::HTTPError
                result = :error
              when Net::HTTPFound
                redirects = args[:redirects] || []
                redirects << url
                location = response["location"]
                if not location =~ /^http.*/
                  location = "#{uri.scheme}://#{uri.host}:#{uri.port}#{location}"
                end
                result = download_image(location, path, :redirects => redirects)
              else
                debug "#{url} -> #{path}"
                if @verbose
                  response.each_header do |k, v|
                    puts "#{k}: #{v}"
                  end
                  puts
                end
                
                # detects weird 404 or 504 responses included in body but not HTTP status
                content_type = response["content-type"]
                if not (content_type =~ /text\/\w+/).nil?
                  error "unexpected content type: #{content_type}"
                  result = :error
                else
                  open(path, 'w') do |out|
                    out << response.body
                  end
                  result = :success
                end
            end
          end
        rescue Exception => e
          error e
          result = :error
        end
        
        return result
      end
      
      def handle_error
        raise Exception("block required") if not block_given?
        begin
          block_result = yield
          result = [:success, block_result]
        rescue
          result = [:error, $!]
        end
        return result
      end
      
  end
  
end